# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from ._bp2pfull import BrokenPowerlawTwoPeakFull as BrokenPowerlawTwoPeakFull
from ._bp2pmultispinfull import (
    BrokenPowerlawTwoPeakMultiSpinMultiTilt as BrokenPowerlawTwoPeakMultiSpinMultiTilt,
    BrokenPowerlawTwoPeakMultiSpinMultiTiltFull as BrokenPowerlawTwoPeakMultiSpinMultiTiltFull,
)
from ._npowerlawmgaussian import NPowerlawMGaussian as NPowerlawMGaussian
from ._powerlawpeak import PowerlawPeak as PowerlawPeak
